# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from datetime import date

project = "rtpengine"
copyright = str(date.today().year) + ", Sipwise"
author = "Sipwise"
release = "master"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "sphinx_rtd_theme"]
myst_heading_anchors = 3

templates_path = ["_templates"]
exclude_patterns = ["_build"]

master_doc = "index"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
htmlhelp_basename = "rtpenginedoc"

# Prevent double dashes -- from becoming single em dashes
smartquotes_action = "qe"
